#include "alphabet.h"

// ===== functions to reduce/enlarge alphabet (no squeezing in ktuples)

// map from input alphabet -> restricted alphabet
static int alphabet_map[SIZE_OF_ALPHABET];
// map from input restritcted alphabet -> alphabet
static int restricted_unmap[SIZE_OF_ALPHABET];


// compute frequencies of symbols in b[n]
// and store them in freq (which should be of len SIZE_OF_ALPHABET
void init_freq(symbol *b, customInt n, customInt *freq)
{
  for(int i=0;i<SIZE_OF_ALPHABET;i++)
    freq[i]=0;
  for(customInt i=0;i<n;i++)
    freq[b[i]]++;  
}

// compute frequencies of symbols in b[n] applying ktuple_decode0
// and store them in freq (which should be of len SIZE_OF_ALPHABET
void ktuple_init_freq(symbol *b, customInt n, customInt *freq)
{
  for(int i=0;i<SIZE_OF_ALPHABET;i++)
    freq[i]=0;
  for(customInt i=0;i<n;i++)
    freq[ktuple_decode0(b[i])]++;  
}



/**
 * Compute alphabet map and its inverse for a string b[n]
 * \return   number of distinct symbols in b
 * */ 
int init_alpha_maps(symbol *b, customInt n)
{
  // init maps with illegal value -1
  for(int i=0;i<SIZE_OF_ALPHABET;i++)
    alphabet_map[i]=restricted_unmap[i] = ILLEGAL_SYMBOL;

  // compute frequencies of the input symbols
  customInt freq[SIZE_OF_ALPHABET];
  init_freq(b,n,freq); // complete scan of b[0..n]
  // compute maps
  int j=0;
  for(int i=0;i<SIZE_OF_ALPHABET;i++)
    if(freq[i]!=0) {
      alphabet_map[i] = j;
      restricted_unmap[j++] = i;
    }
  return j;
}

int alpha_reduce(int c)
{
  assert(alphabet_map[c]!= ILLEGAL_SYMBOL);
  return alphabet_map[c];
}

int alpha_enlarge(int c)
{
  assert(restricted_unmap[c]!= ILLEGAL_SYMBOL);
  return restricted_unmap[c];
}

// remap the string b according to alpha_reduce
// if freq!=null update freq[] according to the remapped values
// freq[] must be appropriately allocated and initialized
void remap_string(symbol *b, customInt n, customInt *freq)
{
  for(customInt i=0;i<n;i++) {
    b[i] = alpha_reduce(b[i]);
    if(freq) freq[b[i]]++;
  }  
}



// ==== functions to handle ktuples containing k chars squeezed together inside a symbol 

// return the number of chars we can fit in one symbol
// assuming chars are in the range 0..n-1
int squeezable(int n)
{
  assert(n<=SIZE_OF_ALPHABET);
  int c=1;
  uint64_t m=n;
  while(m<=SIZE_OF_ALPHABET) { // m = n^c 
    m *=n;
    c++;
  }
  return c-1; 
}

// integer power function (inefficient and only for small values)
int intpow(int b, int e)
{
  assert(e>=0);
  int r=1, i=0;
  while(i<e) {  // r = b^i
    r*=b; i++;
  }
  return r;
}


// constants used to handle k-tuples
static int ktuple_a=0; // original alphabet size
static int ktuple_k;   // # of chars in one symbol 
static int ktuple_ak1; // ktuple_pow[i] = ktuple_a^i 

// init ktuple_* constants 
void ktuple_init(int a, int k)
{
  assert(k<32 && k>0);
  assert(intpow(a,k) <= SIZE_OF_ALPHABET);
  ktuple_a = a;
  ktuple_k = k;
  ktuple_ak1 = intpow(a,k-1);
}
  

// ----------- these comments are no longer valid, see below ------

// since in the input text there is at most one 0
// by construction ktuples can only be of the form 
// a00 ab0 abc bc0 c00 with  a,b,c!=0   
// Note: for k-tuple creation, t=abcde0 becomes 00abcde00
// with k-1 trailing/leading 0's added to create the first and last k-tuple

// We distinguish beween full tuples (eg. cde def) head0 tuple (000 00a 0ab) and
// tail0 tuples (yz0 z0 000). Note there are k tail0 tuples, k-1 head0 tuples
// and n-k ful tuples. In the F column there are only full tuples and 
// tail0 tuples. In the bwt there are only full tuples and head0 tuples.

// when we do the merging, the position in the F-merged-column of the tail0 
// ktuples is computed in advance based on their value and the id of the string
// they belong so they do not have to be moved, we just use them to induce
// the position of other ktuples.

// Simmetrically, when we find a head0 tuple we just have to move it to the 
// correct position but we do not use it to induce the rank of some other
// ktuple. 

// When no squeezing is used, the above considerations apply to the 0 symbol alone


// --------- current situation --------

// when working with many small sequences the overhead of 
// storing the number of occ for each symbol for each sequence can be substantial
// hence this is done only with the -a (small alphabet) command line option 
// This information is used only to init the merge and B arrays so when 
// -a is not present we do not store the occ but recompute them from the bwt
// When also squeezing is used, we msu store oll the infornation about tail0/head0
// tuples in the BWT. To this end given the text abcdefg0 we consider the text 
// ciclically, eg for k=4: abcd bcde cdef defg efg0 fg0a g0ab 0abc  


// return true if s is a ktuple with leading 0s (no longer used)
#if 0
bool ktuple_head0(int s)
{
  return s<ktuple_ak1;
}
#endif

// return true if s is a ktuple with trailing 0s 
// not that such ktuples are never written to the bwt, but they result in the 
// freq arrays, they can be obtained from the bwt using ktuple_decode0 
bool ktuple_tail0(int s)
{
  return (s%ktuple_a)==0;
}

// return true if one of the chars in s is 0
bool ktuple_has0(int s)
{
  for(int i=0;i<ktuple_k;i++) {
    if(s%ktuple_a==0) return true;
    s = s/ktuple_a;
  }
  return false;
}

// decode a tuple removing the symbols after a 0
// given tuple containing a zero, eg ab0cd set to zero
// everything after the 0, ie ab0cd-> ab000, This is done since
// the k tuples with tailing zeros are encoded inside the ktuples with
// leading zeros. ie ab0cd represents both 000cd and ab000
int ktuple_decode0(int s)
{
  int v = s, w = v, mol=1;
  for(int i=0;i<ktuple_k;i++) {
    int c = v%ktuple_a; // last char in v
    if(c==0) return w;  // 0 found 
    // delete c from w
    w = w - c*mol;
    // update v and mol
    v /= ktuple_a; mol *= ktuple_a;
  }
  return s;
}
  

// for a tail0 ktuple return the number of initial nonzero chars
int ktuple_headlen(int s)
{
  assert(ktuple_tail0(s));
  int i, pow = ktuple_ak1;
  for(i=0;i<ktuple_k;i++) {
    if(s/pow==0) break;
    s -= (s/pow)*pow;
    pow /= ktuple_a;
  }
  assert(i<ktuple_k);  
  return i;
}

// return the last item in the ktuple (to create output file) 
int ktuple_last(int s)
{
  return s%ktuple_a;
}

// show glocal values
void ktuple_info(FILE *f) {
  fprintf(f,"Original alphabet size: %d\n",ktuple_a);
  fprintf(f,"Symbols squeezed per char: %d\n",ktuple_k);
  // fprintf(f,"a^%d=%d\n",ktuple_k-1,ktuple_ak1);
}

//display characters of ktuple 
void ktuple_print(FILE *f, int s) {
  int pow = ktuple_ak1;
  for(int i=0;i<ktuple_k;i++) {
    fprintf(f,"%d ",s/pow);
    s -= (s/pow)*pow;
    pow /= ktuple_a;
  }
  fprintf(f,"\n");
}


// return lcp between two distinct ktuples x, y with possibly trailing 0's. 
// Use the rule that 0 chars are always different 
int ktuple_lcp(int x, int y)
{
  int lcp=0;
  int pow = ktuple_ak1;
  while(x/pow!=0 && y/pow!=0 && x/pow == y/pow) { // search leftmost difference between x and y
    assert(pow>=ktuple_a); // if it were x/1==y/1 we would have x==y against our assumption
    x = x%pow; y=y%pow;
    pow /= ktuple_a;       // a^i -> a^{i-1}
    lcp++;
  }
  assert(lcp<ktuple_k);
  return lcp;
}


/**
   squeeze all groups of k char in [0..a-1] in the text whose BWT is bwt[0..n-1] 
   into a single symbol. Note bwt[] is over [0..SIZE_OF_ALPHABET-1] and
   ktuple_a and ktuple_k have been already computed 
   \param [in]   bwt[]  input bwt in the original alphabet
   \param        n      bwt length (for both input and output)
   \param        a      size of the remapped alphabet
   \param        k      chars fitting in one symbol     
   \param [out]  bwt[i] is ktuple preceding suffix sa[i] 
   \param [out]  freq[] frequencies of ktuples in column F after squeezing
                        (if freq!=NULL). If !=NULL freq must have been properly 
                        allocated and initialized to 0

   To squeeze symbols into ktuple the function: 
     1. remap bwt[] in [0..ktuple_a-1]
     2. compute prec array to invert the bwt  
     3. logically reconstruct t[0..n-1] starting with t[n-1]=0 and going
        backward. We maintain in s the ktuple with the last k-chars squeezed 
        together. We write s in the bwt position corresponding
        to the suffix immediately preceeding the ktuple. We also update freq[]
        so that at the end freq[s] stores how many times the ktuple s prefix 
        one of the rows of the suffix array (assuming additional k-1 0's are
        present after t[n-1]=0.
* */
void ktuple_remap_string(symbol *bwt, customInt n, customInt *freq, int a, int k)
{
  // init a^k and a^{k-1}
  assert(n>k);
  assert(k>1);
  int ak1 = intpow(a,k-1);
  int ak = ak1*a;
    
  // remap symbols in [0..a-1] and build F column 
  customInt fcol[SIZE_OF_ALPHABET] = {0};   // init all fcol[] to 0
  remap_string(bwt,n,fcol);  // remap (=reduce) alphabet and write freq to fcol 
  customInt i, newtot=n+1, tot=0;
  for(i=0;i<a;i++) {         // compute F column    
    newtot = tot+ fcol[i];
    fcol[i] = tot;
    tot = newtot; 
  }
  assert(newtot==n);
  if(fcol[1]!=1){  // bwt[] must be a single bwt, not the merging of many 
    printf(CUSTOM_FORMAT" bwts found in a input bwt. Squeezing is currently not supported. Exiting...\n",fcol[1]);
    exit(1); 
  }
  
  // invert BWT computing prec array, write to eos the position of 0 in bwt
  customInt eos=n+1, prec[n];
  for(i=0;i<n;i++) {
    prec[i] = fcol[bwt[i]]++;
    if(bwt[i]==0) eos = i;
  }
  assert(eos<n+1);      // we found eos
  assert(prec[eos]==0); // eos points to first row
  
  // fill freq with the number of occurrences of the tuples 
  // and write to bwt[] the k-tuples preceding each suffix
  customInt pos[k];      // buffer containing the k last seen row indexes
  customInt p=0;         // current pos 
  int s=0;               // current symbol (k 0s starting with t[n-1])
  for(int i=0;i<n;i++) { // invariant: s is the k-tuple prefixing row p
    assert(s<ak);
    if(freq) freq[s]++;     // update k-tuple freq
    if(i>=k) {
      assert(s!=0);
      assert(!ktuple_has0(s));
      assert(bwt[pos[i%k]]==ktuple_last(s));
      bwt[pos[i%k]] = s;    // write ktuple s in front of row seen k iters before
    }
    pos[i%k] = p;           // save current row
    assert(bwt[p]<a);       // make sure this bwt entry is untouched 
    s = s/a + bwt[p]* ak1;  // update ktuple s adding bwt[p] in front 
    p = prec[p];            // update row number moving backward in the test
  }
  assert(p==0);             // we have traversed all rows
  int tail = bwt[0];        // last k symbols in T
  for(i=n;i<n+k;i++) { // write the incomplete symbols
    // ktuple_print(stdout,s);
    assert(bwt[pos[i%k]]==ktuple_last(s));
    bwt[pos[i%k]] = s;
    s = s/a + (tail%a)* ak1;  // update s ciclically
    tail = tail/a;            // lastchar of tail used  
  }
  assert(tail==0);
  assert(s==bwt[0]);
  // additional check
  #ifndef NDEBUG
  if(freq) {
    tot=0;
    for(i=0;i<ak;i++) tot += freq[i];
    assert(tot==n);
    customInt cnt[SIZE_OF_ALPHABET] = {0};
    for(i=0;i<n;i++)
      cnt[ktuple_decode0(bwt[i])]++;
    for(i=0;i<ak;i++) assert(cnt[i] == freq[i]);
  }
  #endif
}

