/* ******************************************************
 * Space economical implementation of irrelevant blocks
 * 
 * TO BE CHANGED
 * blocks can be of two kinds: liquid or solid 
 * when a block is first created it is liquid and becomes solid 
 * immediately after a skip operation provided its size is greater
 * than the constant Small_block_limit 
 * 
 * In a liquid block, occ==NULL so skipping consists in scanning
 * the Z (merge) array exactly as it is done for non-irrelevant blocks
 * but without searching for bocks inside it. 
 * 
 * The rationale is that a block created in the previous iteration
 * has to be processed as an irrelevant block (to avoid the copy newZ -> Z). 
 * In addition if the block is small it is a waste of space (and maybe time)
 * to have the occ array (which has size sizeOfAlphabet + numberOfBWTs)
 * 
 * If occ!=NULL the block is solid and the array occ[] contains
 * in the first numberOfBWTs positions the occurrences of each bwt,
 * and in the following sizeOfAlphabet postions the occurrences of each char 
 * 
 * Blocks can be merged if they are of the same type (liquid liquid or solid solid), or
 * if a liquid block follows a solid block. If a liquid block is followed by a solid block, it
 * must be first solidified and then merged.
 *
 * The main functions in this file are:
 * - proto_add: it is called when  block is first discovered: it merges it to a
 * previous adjacent liquid block, if the latter exists, or else creates a new standalone liquid block
 * - block_process: it is called when an existing block is encountered; it skips it (in case of a solid block)
 * or processes it (in case of a liquid block), trying to merge it to a previous block; if the
 * block is large enough it solidifies it. Unique exception: if it is a liquid block followed by a solid block 
 * to which it could be merged, it solidifies it so that next the two can be merged
 * block_process calls block_skip_solid, block_processAndMergeToSolid, block_processAndSolidify or
 * block_process_liquid, as required.
 *
 * Solid blocks have an occ array of smallSolidInt until their dimension is <= SMALLSOLID_LIMIT
 * When blocks become bigger, occ arrays are inflated to an array of customInt.
 */

// irrelevant solid block to be skipped at each occurrence
typedef struct block {
  customInt beginsAt;
  customInt endsAt;
  union {
    customInt *occ;    // [0..numberOfBWTs-1] [numberOfBWTs .. ]
    smallSolidInt *smallOcc;
  };
  struct block *nextBlock;
} solidBlock;


// a liquid Block is a merge of monochrome blocks that is made solid if 
// it reaches size solid_limit. It it touches a solid block  
// it is merged to it 
typedef struct liqblock {
  customInt beginsAt;
  customInt endsAt;   // only used in assertions 
  int numBwt;
  int sizeOfAlpha;
  int occ_size;
  customInt *occ;    // [0..numberOfBWTs-1] [numberOfBWTs .. ]   
  int solid_limit;   // must be reached to create a solid block    
  bool empty;
  solidBlock *solidList;
  customInt *occList;
  smallSolidInt *smallOccList;
} liquidBlock;


// protoBlock: will become a block if turns out to be monocrhome
typedef struct {
  customInt beginsAt;
  customInt endsAt;
  symbol *start;    // occ of alpha symbol 
  bool mono;
  int color;
} protoBlock;


// --- debug functions

// scan the list of irrelevant blocks and make sure there are no mergeable
// blocks. If verbose report the number of blocks by type 
// NOTE: this is essentially a development/debug function
customInt checkBlocks(solidBlock *b, customInt pfx, g_data *g, bool lastRound)
{
  customInt tot=0, mergeable=0, small_solid=0;

  while(b!=NULL) {
    tot++;
    if(b->nextBlock!=NULL && b->endsAt == b->nextBlock->beginsAt) 
      mergeable++; 
    if(b->endsAt-b->beginsAt<= SMALLSOLID_LIMIT) small_solid++;
    b = b->nextBlock;  // check what comes after
  }
  if(g->verbose>2 || (g->verbose>1 && lastRound)) 
    printf("Lcp: "CUSTOM_FORMAT"  Blocks: "CUSTOM_FORMAT" total, "CUSTOM_FORMAT" small_solid\n", pfx-1, tot,small_solid);
  if(mergeable!=0)
    printf("Lcp: "CUSTOM_FORMAT"  Fatal error: "CUSTOM_FORMAT" mergeable blocks found\n",pfx-1,mergeable);
  return mergeable; 
}


// debug only
void showIrrelevants(solidBlock *b)
{
  printf("Irrelevants: ");
  while(b!=NULL) {
    assert(b->occ!=NULL);
    if (b->endsAt - b->beginsAt <= SMALLSOLID_LIMIT) 
      printf("("CUSTOM_FORMAT" "CUSTOM_FORMAT") ",b->beginsAt,b->endsAt);
    else 
      printf("["CUSTOM_FORMAT" "CUSTOM_FORMAT"] ",b->beginsAt,b->endsAt);
    b = b->nextBlock;
  }
  puts("END");
}

// debug only: show occ's
void showBlock(solidBlock *b, g_data *g, int num)
{
  printf("Beg "CUSTOM_FORMAT" End "CUSTOM_FORMAT" ",b->beginsAt, b->endsAt);
  if(b->endsAt - b->beginsAt > SMALLSOLID_LIMIT) {
    printf("occ: ");
    for(int i=0;i<num;i++) printf(CUSTOM_FORMAT" ",b->occ[i]);
  } else {
    printf("smallOcc: ");
    for(int i=0;i<num;i++) printf(SMALLSOLID_FORMAT" ",b->smallOcc[i]);
  }
  puts("SolEND");
}

void showLiquid(liquidBlock *b)
{
  printf("Beg "CUSTOM_FORMAT" End "CUSTOM_FORMAT" ",b->beginsAt, b->endsAt);
  customInt totA=0, totB=0;
  for(int i=0;i<b->numBwt;i++) totB += b->occ[i];
  for(int i=0;i<b->sizeOfAlpha;i++) totA += b->occ[b->numBwt+i];
  printf("Chars: "CUSTOM_FORMAT" Bwts: "CUSTOM_FORMAT" Empty: %d ",totA,totB,b->empty);  
  puts("LiqEND");
}



// --- alloc/dealloc functions

// create an empty liquid block, called once for iteration 
liquidBlock *liquid_new(g_data *g){
  liquidBlock *newB = malloc(sizeof(*newB));
  if(newB==NULL) die("Out of mem in liquid_new()");
  newB->beginsAt = 0; 
  newB->endsAt = 0; 
  newB->empty = true; 
  newB->occ_size = g->numBwt+g->sizeOfAlpha;
  newB->numBwt = g->numBwt;
  newB->sizeOfAlpha = g->sizeOfAlpha;
  newB->solid_limit = max(g->solid_limit,newB->occ_size);
  // solid limit cannot be smaller than mergeLen  
  newB->solid_limit = min(newB->solid_limit,g->mergeLen);  
  newB->occ = calloc(newB->occ_size,sizeof(customInt));
  if(newB->occ==NULL) die("Out of mem in liquid_new()");
  newB->solidList = NULL;
  newB->occList = NULL;
  newB->smallOccList = NULL;
  return newB;
}

// free liquid block, called once for iteration 
void liquid_free(liquidBlock *b)
{
  // free list of solid blocks
  assert(b!=NULL && b->occ != NULL);
  {solidBlock *next=b->solidList;
  while(next!=NULL) {
    solidBlock *tmp = next->nextBlock;
    free(next); next = tmp;
  }}
  // free list of occList
  {customInt *next=b->occList;
  while(next!=NULL) {
    customInt **n = (customInt **) next; // pretend next is an array of pointers customInt *
    customInt *tmp = n[0];
    free(next); next = tmp;
  }}
  // free list of smallOccList
  {smallSolidInt *next= b->smallOccList;
  while(next!=NULL) {
    smallSolidInt **n = (smallSolidInt **) next; // pretend next is an array of pointers customInt *
    smallSolidInt *tmp = n[0];
    free(next); next = tmp;
  }}
  free(b->occ);
  free(b);
}

// get an occ list: if possible from s->occList otherwise via malloc 
customInt *get_occ(liquidBlock *s) {
  customInt *occ = s->occList;  // try occlist
  if(occ == NULL) {
    occ = malloc(s->occ_size* sizeof(customInt));
    if(occ==NULL) die("Out of mem in get_occ()");
  }
  else { // update list occList
    customInt **n = (customInt **) occ;
    s->occList = n[0];
  }
  return occ;  
}

// get a  smallOcc list: if possible from s->occList otherwise via malloc 
smallSolidInt *get_smallocc(liquidBlock *s) {
  smallSolidInt *occ = s->smallOccList;  // try smallOcclist
  if(occ == NULL) {
    occ = malloc(s->occ_size* sizeof(smallSolidInt));
    if(occ==NULL) die("Out of mem in get_smallocc()");
  }
  else { // update list occList
    smallSolidInt **n = (smallSolidInt **) occ;
    s->smallOccList = n[0];
  }
  return occ;  
}


// add to s->occList the array occ
void free_occ(customInt *occ, liquidBlock *s) {
    customInt **n = (customInt **) occ; // pretend occ is an array of pointers
    n[0] = s->occList;
    s->occList = occ;
}

// add to s->occList the array occ
void free_smallocc(smallSolidInt *occ, liquidBlock *s) {
    smallSolidInt **n = (smallSolidInt **) occ; // pretend occ is an array of pointers
    n[0] = s->smallOccList;
    s->smallOccList = occ;
}



// free a solid block (called only after a three way merge merge_sls)
// saving its space to x->solidList;
void block_free(solidBlock *b, liquidBlock *x)
{
  assert(b!=NULL && b->occ != NULL);
  // free occ's
  if(b->endsAt -b->beginsAt > SMALLSOLID_LIMIT)
    free_occ(b->occ,x);
  else // smallOcc  
    free_smallocc(b->smallOcc,x);
  // free block itself  
  b->nextBlock = x->solidList;
  x->solidList = b;
}

// free all mem in  list of solid blocks 
void solidList_free(solidBlock *b, liquidBlock *x) {
  while(b!=NULL) {
    solidBlock *tmp = b->nextBlock;
    block_free(b,x);
    b = tmp;
  }
}
  

/* * substitutes the smallOcc array with a customInt array
 * \param b  pointer to block to be processed
 * \param s: size occ array 
 * */
static void block_inflate(solidBlock *b, liquidBlock *s) {
  assert(b->endsAt - b->beginsAt <= SMALLSOLID_LIMIT);
  customInt *bigOcc = get_occ(s);
  for (int i = 0; i < s->occ_size; ++i) {
    bigOcc[i] = (customInt) b->smallOcc[i];
  }
  free_smallocc(b->smallOcc,s);  
  b->occ = bigOcc;
}


/* * creates a new block to store data from the liquid block s
 *  \param s pointer to liquid block
 * 
 * Note: this is the only function where solid blocks are created
 * */
static solidBlock *liquidToSolid(liquidBlock *s) {
  assert(s->endsAt - s->beginsAt >= s->solid_limit);
  
  // ----- get solid block from s->solidList or malloc
  solidBlock *newB = s->solidList;
  if(newB!=NULL) 
    s->solidList = newB->nextBlock;
  else {
    newB = malloc(sizeof(*newB));
    if(newB==NULL) die("Out of mem in liquidToSolid()");
  }
  // --- init block size
  newB->beginsAt = s->beginsAt;
  newB->endsAt = s->endsAt;
  // --- get and init occ/smallOcc array
  if (newB->endsAt - newB->beginsAt <= SMALLSOLID_LIMIT) {
    // small solid block  try get mem from s->occList
    newB->smallOcc = get_smallocc(s);
    for (int i = 0; i < s->occ_size; ++i)  // copy occ from s
      newB->smallOcc[i] = (smallSolidInt) s->occ[i];
  } 
  else { // large solid block  try get mem from s->occList
    newB->occ = get_occ(s);
    for (int i = 0; i < s->occ_size; ++i) // copy occ from s
      newB->occ[i] = s->occ[i];
  }
  // clear liquid block 
  s->empty=true;
  memset(s->occ, 0, (s->occ_size) * sizeof(customInt));
  return newB;
}



// --------- single block functions


/* * skips an irrelevant block
 * \param this  pointer to block to be skipped
 * \param g:    all general data
 * */
static void skip(solidBlock *this, g_data *g) {
  if(this->endsAt - this->beginsAt <= SMALLSOLID_LIMIT){ 
    for (int col = 0; col < g->numBwt; col++) 
      g->inCnt[col] += this->smallOcc[col];    //skipping this
    for (int c = 0; c < g->sizeOfAlpha; ++c) 
      g->F[c]+= this->smallOcc[g->numBwt+c];
  } 
  else { 
    for (int col = 0; col < g->numBwt; col++) 
      g->inCnt[col] += this->occ[col];        //skipping this
    for (int c = 0; c < g->sizeOfAlpha; ++c) 
      g->F[c] += this->occ[g->numBwt+c];
  }
}

// add proto block to the current liquid block
static void add_proto2liquid(protoBlock *proto,liquidBlock *liquid)
{
  assert(liquid->endsAt==proto->beginsAt);
  assert(proto->endsAt>proto->beginsAt);
  liquid->endsAt=proto->endsAt;
  // update color occurrences 
  liquid->occ[proto->color] += (proto->endsAt-proto->beginsAt);
  // update alpha occurences 
  for(int i=0;i<proto->endsAt-proto->beginsAt;i++)
    liquid->occ[liquid->numBwt + (proto->start)[i]] +=1;
  liquid->empty = false;
}


// ---- merge functions 

// merge liquid block b to solid s
// b is cleared after the merging 
static void merge_liquid(liquidBlock *b, solidBlock *s) {
  assert(!b->empty);
  // --- merge occs. 
  if((s->endsAt - s->beginsAt) + (b->endsAt - b->beginsAt) <= SMALLSOLID_LIMIT){
    for (int c = 0; c < b->occ_size; ++c) // merged block is small 
      s->smallOcc[c] += (smallSolidInt) b->occ[c];
  } 
  else {
    if((s->endsAt - s->beginsAt)<= SMALLSOLID_LIMIT)
      block_inflate(s,b);       // enlarge if required 
    for (int c = 0; c < b->occ_size; ++c) 
      s->occ[c] += b->occ[c];
  }
  // --- update extremes
  if(b->endsAt==s->beginsAt)
    s->beginsAt = b->beginsAt;  // liquid before solid
  else {
    assert(s->endsAt==b->beginsAt); // solid before liquid
    s->endsAt = b->endsAt;
  }
  // --- clear b
  b->empty = true;
  memset(b->occ, 0, (b->occ_size) * sizeof(customInt));
}


// check if liquid block b can be merged to last or made a solid by itself
// here is where solid_limit is used  
// in any case b is emptied
solidBlock *finalize_liquid(solidBlock *last, liquidBlock *b, solidBlock *next)
{
  assert(!b->empty);
  if(last!=NULL && last->endsAt==b->beginsAt) {
    // merge b into last and empty it 
    merge_liquid(b,last);
  }
  else if(b->endsAt-b->beginsAt>= b->solid_limit) {
    // create new solid and empty b which becomes last block    
    solidBlock *s = liquidToSolid(b);
    s->nextBlock = next;
    if(last!=NULL) last->nextBlock = s;
    last = s;
  }
  else { // nothing to do, just empty b
    b->empty = true;
    memset(b->occ, 0, (b->occ_size) * sizeof(customInt));
  }
  assert(b->empty);
  return last;
}


// merge solid s to liquid b to solid z
// b is cleared after the merging. z is dellocated 
static void merge_sls(solidBlock *s, liquidBlock *b, solidBlock *z) {
  // showBlock(s,NULL,0); showBlock(z,NULL,0); showLiquid(b);
  assert(s && z && b && !b->empty);
  assert(s->endsAt==b->beginsAt && b->endsAt==z->beginsAt);
  customInt ssize = s->endsAt - s->beginsAt;
  customInt zsize = z->endsAt - z->beginsAt;
  bool zreduced = false;
  // --- merge occs 
  if(ssize + zsize + (b->endsAt - b->beginsAt) <= SMALLSOLID_LIMIT){
    for (int c = 0; c < b->occ_size; ++c) // all blocks are small 
      s->smallOcc[c] += z->smallOcc[c] + (smallSolidInt) b->occ[c];
  } 
  else if(ssize>SMALLSOLID_LIMIT && zsize>SMALLSOLID_LIMIT) {
      for (int c = 0; c < b->occ_size; ++c) // all blocks are large  
        s->occ[c] += z->occ[c] + b->occ[c];
  }
  else { // mixed small/large blocks
    if(zsize>SMALLSOLID_LIMIT) {// s must be  small, swap with z
      customInt *tmp = z->occ; z->smallOcc = s->smallOcc; s->occ = tmp; zreduced=true;
    }
    else {  // z is small, s can be small or large 
      if(ssize<=SMALLSOLID_LIMIT)  // if s is small enlarge it
        block_inflate(s,b);
    }
    // now s is large and z is small
    for (int c = 0; c < b->occ_size; ++c) 
        s->occ[c] += z->smallOcc[c] + b->occ[c];
  }
  // ----  update endpoints and nextBlock
  s->endsAt = z->endsAt;
  s->nextBlock = z->nextBlock;
  // --- clear b and free z
  b->empty = true;
  memset(b->occ, 0, (b->occ_size) * sizeof(customInt));
  if(zreduced) z->endsAt = z->beginsAt; // when a large z we gave a a smallocc from s 
  block_free(z,b);
}


// functions handling the two-bit array used in lieu of B
// for bwt-only gap algorithm. 

// alloc and set to zero a two-bit array
uint64_t *tba_alloc(customInt n, bool mmapB)
{
  uint64_t *a;
  customInt j = (n+31)/32;
  if(mmapB) {
    a = mmap(NULL,j*sizeof(*a),PROT_READ|PROT_WRITE,MAP_PRIVATE|MAP_ANONYMOUS,-1,0);
    if(a== MAP_FAILED) die(__func__);
  }
  else {
    a = calloc(j,sizeof(*a));
    if(a==NULL) die(__func__);
  }
  return a;
}

void tba_free(uint64_t *a, customInt n, bool mmap)
{
  if(mmap) {
    customInt j = (n+31)/32;
    int e = munmap(a,j*sizeof(*a));
    if(e!=0) die(__func__);
  }
  else free(a);
}
// test a[i]==0 (not used)
inline bool tba_is_zero(uint64_t *a,customInt i)
{
  customInt q = i/32;
  customInt r = i%32;
  uint64_t mask = (uint64_t) 3 << (2*r);
  return (a[q] & mask)==0;
}
// test a[i]==11 (not used)
inline bool tba_is_11(uint64_t *a,customInt i)
{
  customInt q = i/32;
  customInt r = i%32;
  return ((a[q] >> (2*r)) & 3) ==3; // shift and check last 2 bits 
} 
// set a[i] = a[i] | m
inline void tba_or_m(uint64_t *a, customInt i, uint64_t m)
{
  assert(m==1 || m==2);
  customInt q = i/32;
  customInt r = i%32;
  a[q] = a[q] | (m<< (2*r));
}
inline void tba_mark_if0(uint64_t *a, customInt i, uint64_t m)
{
  assert(m==1 || m==2);
  customInt q = i/32;
  customInt r = i%32;
  if( ((a[q]>> (2*r)) & 3) == 0)
    a[q] = a[q] | (m<< (2*r));
}
// return true if we are at the begining of a block 
// that is if a[i]==11 || a[i]== ~m
// if a[i]==11 set recent to false, 
// if a[i]==~m set recent to true and set a[i] to 11 (for next iterations)
// otherwise return false and do not change recent  
inline bool tba_block_test_set(uint64_t *a,customInt i, int m, bool *recent)
{
  assert(m==1 || m==2);
  customInt q = i/32;
  customInt r = i%32;
  int b2 = (a[q] >> (2*r)) & 3;       // shift and check last 2 bits
  if(b2==3-m) {
    *recent = true;
    a[q] = a[q] | ( (uint64_t) 3 << (2*r)); // make the block not recent 
    return true;
  }
  else if(b2==3) {
    *recent = false;
    return true;
  }
  return false; // do not change recent 
} 
// report value (not used)
int tba_get(uint64_t *a,customInt i)
{
  customInt q = i/32;
  customInt r = i%32;
  int b2 = (a[q] >> (2*r)) & 3;       // shift and check last 2 bits
  return b2;
}

// ---- same tba functions for the case the B array stored in the two most signficat bits of the merge array 

// set a[i] = a[i] | m
inline void tba_or_m8(uint8_t *a, customInt i, uint64_t m)
{
  assert(m==1 || m==2);
  a[i] = a[i] | (m<< 6);
}
inline void tba_mark8_if0(uint8_t *a, customInt i, uint64_t m)
{
  assert(m==1 || m==2);
  if( (a[i] & 0xC0) == 0)   // if 2 msb are 0
    a[i] = a[i] | (m<< 6);  // write m on them 
}
// return true if we are at the begining of a block 
// that is if a[i]==11 || a[i]== ~m
// if a[i]==11 set recent to false, 
// if a[i]==~m set recent to true and set a[i] to 11 (for next iterations)
// otherwise return false and do not change recent  
inline bool tba_block_test_set8(uint8_t *a,customInt i, int m, bool *recent)
{
  assert(m==1 || m==2);
  int b2 = (a[i] >> 6) & 3;       // shift and check 2 msb 
  if(b2==3-m) {
    *recent = true;
    a[i] = a[i] | ( 3 << 6); // make the block not recent 
    return true;
  }
  else if(b2==3) {
    *recent = false;
    return true;
  }
  return false; // do not change recent 
} 
