#ifndef CONFIG_H
#define CONFIG_H
#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <stdbool.h>
#include <errno.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <time.h>
#include <pthread.h>
#include <semaphore.h>
#include <inttypes.h>

// size of buffer for building file names
#define Filename_size 500
// size of all buffers for producer/consumer threads
#define Threads_buf_size 1

// type used to represent an input symbol
typedef unsigned char symbol;
#define SIZE_OF_ALPHABET 256
#define BWT_EOF 0x00
#define ILLEGAL_SYMBOL (-1)

//type used to represent occurrences in small solid blocks
typedef uint16_t smallSolidInt;
#define SMALLSOLID_FORMAT "%"PRIu16
#define SMALLSOLID_LIMIT UINT16_MAX

// type used to represent positions in the BWT's 
// define the maximum (merged) BWT size 
typedef uint64_t customInt;
#define CUSTOM_FORMAT "%"PRIu64
#define MAX_OUTPUT_SIZE 0xFFFFFFFEULL

// type used to represent LCP values (B array)
#ifndef BSIZE
#define BSIZE 1
#endif
#if BSIZE==1
typedef uint8_t lcpInt;
#define LCP_FORMAT "%"PRIu8
#define MAX_LCP_SIZE 0xFFULL
#define LCP_EXT "1.lcp"
#define HM_LCP_EXT "1.lcpHM"
#elif BSIZE==2
typedef uint16_t lcpInt;
#define LCP_FORMAT "%"PRIu16
#define MAX_LCP_SIZE 0xFFFFULL
#define LCP_EXT "2.lcp"
#define HM_LCP_EXT "2.lcpHM"
#else
typedef uint32_t lcpInt;
#define LCP_FORMAT "%"PRIu32
#define MAX_LCP_SIZE 0xFFFFFFFFULL
#define LCP_EXT "4.lcp"
#define HM_LCP_EXT "4.lcpHM"
#endif

// type used to represent the Merge (aka Z) array
// defines the maximum number of input BWT's for a single round
// Note that, for LCP computation, we need an open file for each BWT, 
// so the max number of BWT can be limited in practice by the maximum 
// number of open files allowed by the system  (eg 1024 on my laptop)
#define ZSIZE 1
#if ZSIZE==1
typedef uint8_t palette;
#define PALETTE_FORMAT "%"PRIu8
#define MAX_NUMBER_OF_BWTS 0x100ULL
#elif ZSIZE==2
typedef uint16_t palette;
#define PALETTE_FORMAT "%"PRIu16
#define MAX_NUMBER_OF_BWTS 0x10000ULL
#else
typedef uint32_t palette;
#define PALETTE_FORMAT "%"PRIu32
#define MAX_NUMBER_OF_BWTS 0x100000000ULL
#endif


#define BWT_EXT "bwt"
#define LEN_EXT "len"
#define HM_BWT_EXT "bwtHM"

 
typedef struct {
  // input: shared among threads
  symbol **bws;            // bws[0] ... bws[numBwt-1] are input bwt
  customInt *bwtLen;       // bwtLen[i] is size of bwt[i]
  customInt **bwtOcc;      // occ of each symbol in each bwt 
  int numBwt;              // number of bwt to be merged
  customInt mergeLen;      // sum_i bwtLen[i]
  customInt symb_offset;   // offset with respect to the global bwt/lcp (used to read in single lcp file)
  int sizeOfAlpha;         // size of the alphabet   
  int charXsymb;           // # of chars in each symbol
  bool bwtOnly;            // do not compute LCP
  bool singleFile;         // lcp are in a single file
  bool lcpMem;             // lcp are stored in memory
  bool smallAlpha;         // the alphabet is small
  bool mmapZ;              // mmap Z arrays
  bool mmapB;              // mmap B array
  bool mmapBWT;            // mmap BWT arrays
  char *lcpinPath;         // base path for lcp files
  lcpInt **lcps;           // if lcpMem, lcps[0] ... lcps[numBwt-1] are input lcp values
  int solid_limit;         // irrelevant blocks become solid after this size
  int verbose;
  union {                   // in Gap use one or the other according to bwtOnly 
    lcpInt *blockBeginsAt;  // B array and then merged LCP. output but allocatd by caller  (shared as well)
    uint64_t *bitB;         // B array represented in 2 bits private 
  };
  // glocal private
  palette *mergeColor;     // Z
  palette *newMergeColor;  // newZ 
  customInt *firstColumn;  // compact first column array
  customInt *F;            // positions inside the newMerge (array F in the pseudocode)   
  customInt *inCnt;        // counters inside each bwt (k_i in the pseudo code)
  symbol *diffChar;        // for sqeeezed lcp computation 
  symbol *newDiffChar;
} g_data; 


// generic data structure for a  1 producer N consumers scheme
typedef struct {
  int pindex;             // index in the buffer for producer
  int cindex;             // index in the buffer for consumers  
  pthread_mutex_t cmutex; // consumer mutex 
  sem_t free_slots;       // free slots in the buffer
  sem_t ready;            // data ready to be processed
  void *buffer;           // generic data buffer
  int buf_size;           // size of buffer
  // from here fields are more or less specific for merge 
  int remaining;          // remaining jobs before a sincronizaion 
  pthread_mutex_t remutex;// mutex for access to remaining
  pthread_cond_t recond;  // condition variable for remaining 
  union {
    bool hm;                // used for hm vs gap
    g_data *data;
  };
} pc_system;




#endif
