// Sample implementation of the GAP algorithm 
// Copyright 2017 Lavina Egidi, Giovanni Manzini
//
#include "util.h"
#include "alphabet.h"
#include "gap.h"
#include "malloc_count/malloc_count.h"

void writeOutputFiles(g_data *g, char *path, int num, bool hm);

// from multiround.c
void remap_bwts(g_data *g, bool squeeze, int num_threads);
void multiround(bool hm, int group_size,char *path, g_data *input, int);


void usage(char *name, g_data *g){
  printf("\nUsage: %s [options] PATH N\n\n",name);
  puts("Merges N BWTs (and optionally LCPs) using the HM or Gap (default) algorithm");
  puts("If the -c option is used the bwt's and lcp's are all in PATH.bwt and PATH.lcp");
  puts("with the length of bwt stored as a 4 byte uint in PATH.len. If -c is not used");
  puts("the names of input bwt/lcp files are obtained appending to PATH the extensions");
  puts("   .0.bwt 1.bwt ...   and   .0.D.lcp 1.D.lcp ...");
  puts("where D is the number of bytes to represent a lcp value chosen at compile time,");
  printf("currently D=%zu.\n\n",sizeof(lcpInt));
  puts("Available options:");
  puts("\t-h    this help message");
  puts("\t-l    merge also lcp values");
  puts("\t-c    input is a single concatenated bwt/lcp + len file");   
  puts("\t-a    assume alphabet is small");   
  printf("\t-g G  max # BWTs merged simultaneously (def. %llu)\n", MAX_NUMBER_OF_BWTS);   
  puts("\t-m    use H&M algorithm");
  printf("\t-s S  minimum solid block size (def %d)\n",g->solid_limit);
  puts("\t-p P  use N parallel threads for merging/squeezing (def 0)");
  puts("\t-T    mmap input BWT arrays");
  puts("\t-L    mmap lcp values");
  puts("\t-Z    mmap merge arrays");
  puts("\t-B    mmap B array");
  puts("\t-z    squeeze input chars in a symbol");
  puts("\t-v    verbose output (more v's for more verbose)\n");
}

int main(int argc, char *argv[]) {
  extern int optind, opterr, optopt;
  extern char *optarg;
  int c, group_size, numBwt;
  char *path;
  g_data g;
   
  // default option values
  g.verbose=0; g.solid_limit = 256;
  group_size=MAX_NUMBER_OF_BWTS;      
  g.bwtOnly = true; g.singleFile=false;
  bool hm = false, squeeze=false;
  g.lcpMem=g.smallAlpha=g.mmapZ=g.mmapBWT=g.mmapB=false;
  int num_threads = 0;
  while ((c=getopt(argc, argv, "vhcalmp:zg:s:ZTBL")) != -1) {
    switch (c) 
      {
      case 'v':
        g.verbose++; break;      // verbose output (can be repeated)
      case 'l':
        g.bwtOnly=false; break;  // compute also lcp array 
      case 'm':
        hm=true; break;        // use hm algorithm 
      case 'z':
        squeeze=true; break;   // squeeze as many chars as posssible in a symbol
      case 'a':
        g.smallAlpha=true; break;   // assume alphabet is small and use bwtOcc[][]
      case 'c':
        g.singleFile=true; break;       // input is a single concatenated BWT/LCP 
      case 'g':
        group_size = atoi(optarg);  // size for multigroup algorithm 
        break;      
      case 's':
        g.solid_limit = atoi(optarg);  // smallest solid block 
        break;                
      case 'p':
        num_threads = atoi(optarg);    // bumber of consumer threads 
        break;       
      case 'Z':
        g.mmapZ=true; break;   // mmap merge and newmerge
      case 'B':
        g.mmapB=true; break;   // mmap B array
      case 'T':
        g.mmapBWT=true; break;   // mmap input BWTs
      case 'L':
        g.lcpMem=true; break;    // store lcp in mmaped memory  
      case 'h':                // usage instruction 
      case '?':
        usage(argv[0],&g);
        exit(EXIT_FAILURE);
      }
  }
  if(num_threads <0) {
    printf("Invalid number of threads, must be non negative\n");
    exit(EXIT_FAILURE);
  }
  if(group_size<2 || group_size>MAX_NUMBER_OF_BWTS) {
    printf("Invalid group size. Must be in range [2,%llu]\n",MAX_NUMBER_OF_BWTS);
    exit(EXIT_FAILURE);
  } 
  if(g.bwtOnly && g.lcpMem) {
    printf("LCP in memory makes sense only if you compute LCPs\n");
    exit(EXIT_FAILURE);
  }
  if(!g.singleFile && g.mmapBWT) {
    printf("BWT can be mmapped only if it is in a single file\n");
    exit(EXIT_FAILURE);
  }
  if(!g.singleFile && g.lcpMem) {
    printf("LCP can be mmapped only if it is in a single file\n");
    exit(EXIT_FAILURE);
  }

  
  if(optind+2==argc) {
    path=argv[optind++];
    numBwt=atoi(argv[optind]);
    if(group_size==0 && numBwt>MAX_NUMBER_OF_BWTS) {
      printf("Cannot merge so many BWTs (%d vs %llu). Use multiround\n",numBwt,MAX_NUMBER_OF_BWTS);
      exit(EXIT_FAILURE);
    }
  }
  else  {
    usage(argv[0],&g);
    exit(EXIT_FAILURE);
  }
  if(!g.bwtOnly && group_size<numBwt && g.singleFile==0 && !g.lcpMem) {
    printf("Multiround LCP computation requires -c or -x\n");
    exit(EXIT_FAILURE);
  }

  if(g.verbose>0) {
    puts("Command line:");
    for(int i=0;i<argc;i++)
      printf(" %s",argv[i]);
    puts("");
  }

  if(g.verbose>1) {
    puts(            "***************************************************");
    if(g.bwtOnly) puts("* Merge of BWTs");
    else          puts("* Merge of BWTs and LCPs");
    if(hm)        puts("* Using multiround HM algorithm");
    else          puts("* Using multiround Gap algorithm");
    printf(          "* Max #BWT %llu, Max LCP: %llu, Max mergeSize: %llu\n",
                     MAX_NUMBER_OF_BWTS, MAX_LCP_SIZE, MAX_OUTPUT_SIZE);  
    puts(            "* Using Timo Bingmann's malloc_count to measure heap usage");
    puts(            "***************************************************");
  }

  // read BWT values
  g.numBwt = numBwt;
  g.bws = (symbol **) malloc (g.numBwt * sizeof(symbol *));
  g.bwtLen = (customInt *) malloc (g.numBwt * sizeof(customInt));
  if(!g.bws || !g.bwtLen) die(__func__);
  if(g.singleFile) 
    g.mergeLen = readBWTsingle(path, &g);   // g.numBwt, bwt, sizeOfBWT);
  else  
    g.mergeLen = readBWTfiles(path, g.numBwt, g.bws, g.bwtLen);

  // init additional g fields 
  g.lcpinPath = path; g.symb_offset = 0;
  g.bwtOcc=NULL; g.sizeOfAlpha = 0;  //these two will be initialized later
  // if requested allocate space for and read lcp Values (B array)
  if(!g.bwtOnly) {
    if(g.mmapB) {
      g.blockBeginsAt = mmap(NULL,g.mergeLen*sizeof(lcpInt),PROT_READ|PROT_WRITE,MAP_PRIVATE|MAP_ANONYMOUS,-1,0);
      if(g.blockBeginsAt== MAP_FAILED) die("main (mmap B array)");
    }
    else {
      g.blockBeginsAt = (lcpInt *) malloc(g.mergeLen*sizeof(lcpInt));
      if(!g.blockBeginsAt) die(__func__);
    }
    if(g.lcpMem) // if requested read lcp values
      initLCPmem(&g);
  }
  else g.blockBeginsAt = NULL;

  // data input completed start measuring time 
  clock_t start = clock();
  time_t start_wc = time(NULL);
  
  // remap BWTs optionally squeeze them, and init g->sizeOfAlpha and
  // is g->smallAlpha is true init also g>bwtOcc[i]  
  remap_bwts(&g,squeeze,num_threads);    
  
  // do the merging
  multiround(hm,group_size,path,&g, num_threads);

  // report running times  
  double elapsed = (clock()-start)/(double)(CLOCKS_PER_SEC);
  double elapsed_wc = difftime(time(NULL),start_wc);
  printf("Elapsed time: %.4lf secs\n", elapsed_wc);
  printf("Computing time: %.4lf secs\n", elapsed);
  printf("Size of merged BWT: "CUSTOM_FORMAT, g.mergeLen);
  printf(" Alphabet size: %d\n", g.sizeOfAlpha);
  printf("Microseconds per symbol: %.4lf\n", elapsed_wc*1000000.0/g.mergeLen);
  printf("Peak memory allocation: %zu bytes, %.4lf bytes/symbol\n",
           malloc_count_peak(), (double)malloc_count_peak()/g.mergeLen);

  // write content of g->bwts[0] and g->blockBeginsAt 
  writeOutputFiles(&g,path,numBwt,hm);

  // free all
  if(!g.bwtOnly) {
    if(g.lcpMem) {
      int e = munmap(g.lcps[0],g.mergeLen*sizeof(lcpInt));
      if(e) die("main (unmap lcps)");
      free(g.lcps); // if used free storage for lcp values
    }
    if(g.mmapB) {
      int e = munmap(g.blockBeginsAt,g.mergeLen*sizeof(lcpInt));
      if(e) die("main (unmap B array)");
    }
    else free(g.blockBeginsAt); // if used free storage for lcp values
  }
  // free BWT stuff
  if(g.smallAlpha) {free(g.bwtOcc[0]); free(g.bwtOcc);}
  free(g.bwtLen);
  if(g.mmapBWT) {
    int e = munmap(g.bws[0],g.mergeLen*sizeof(symbol));
    if(e) die("main (unmap bws)");
  }
  else  
    free(g.bws[0]); // deallocate all bwt's (they are contiguous)
  free(g.bws);
  return 0; // Done!!!
}


/**
 * create output bwt file and write the content of g->bwt[0] (remapped 
 * and unsqueezed) to it 
 * if required create lcp file and write the content of g->blockBeginsAt to it
 * To be changed if one of both bwt/lcp are mmaped files 
 * */
void writeOutputFiles(g_data *g, char *path, int numBwt, bool hm)
{
  char filename[Filename_size];

  // write to output files
  if(g->verbose>1) puts("Writing output files");
  // open bwt output file 
  snprintf(filename,Filename_size,"%s.out.%d.%s",path,numBwt, hm?HM_BWT_EXT:BWT_EXT);
  FILE *f = fopen(filename,"wb");
  if(f==NULL) die(__func__);
  size_t w = fwrite(g->bws[0], sizeof(symbol), g->mergeLen, f);
  if(w!=g->mergeLen) die("Error writing final BWT");
  if(fclose(f)!=0) die(__func__);
  // same for LCP if required
  if(!g->bwtOnly) {
    snprintf(filename,Filename_size,"%s.out.%d.%s",path,numBwt,hm?HM_LCP_EXT:LCP_EXT);
    f = fopen(filename,"wb");
    if(f==NULL) die(__func__);
    w = fwrite(g->blockBeginsAt,sizeof(lcpInt),g->mergeLen,f);
    if(w!=g->mergeLen) die("Error writing final LCP");
    fclose(f);
  }  
}

