/*
 * Induced Suffix Sorting and LCP array construction for String Collections
 *
 * Authors: Felipe A. Louza, Simon Gog, Guilherme P. Telles
 * contact: louza@ic.unicamp.br
 * 
 * version 1.1
 * 01/09/2015
 *
 * Modifications by Giovanni Manzini 12/01/17
 */

#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <assert.h>
#include "lib/file.h"
#include "lib/suffix_array.h"
#include "lib/lcp_array.h"
#include "external/malloc_count/malloc_count.h"
#include "src/gsais.h"
#include "src/gsaca-k.h"


#ifndef DEBUG
#define DEBUG   0
#endif


void usage(char *name){
  printf("\n\tUsage: %s [options] FILE N\n\n",name);
  puts("Computes SA (and optionally LCP array) for the first N sequences of a");
  puts("collection using an algorithm from the gsa-is suite from Louza et al ");
  puts("DCC 16 paper. Sequences from FILE are extracted according to FILE's");
  puts("extension; currently supported extensions are: .txt .fasta .fastq\n");
  puts("Available options:");
  puts("\t-h    this help message");
  puts("\t-m M  algorithm to be used (see Louza et al DCC 16 paper)");
  puts("\t      1=SAIS* 2=SACA* 3=SAIS 4=SACA 5=gSAIS 6=gSAKA (default)");
  puts("\t-l    compute LCP array as well (use only with option -o)");
  puts("\t-c    check SA and LCP");
  puts("\t-o    output SA (ext: .sa) and possibly LCP (ext: .sa_lcp)");
  puts("\t-b    output BWT (ext: .bwt)");
  puts("\t-g D  output LCP in gap format D bytes per entry (ext: .D.lcp)");
  puts("\t-x    extract individual input files and stop");
  puts("\t-X    convert input to raw+len format (ext: .cat .len) and stop");
  puts("\t-L    lengths of the input sequences in FILE.len (no separator)");
  puts("\t-v    verbose output (more v's for more verbose)\n");
  exit(EXIT_FAILURE);
}

/*******************************************************************/

int main(int argc, char** argv){
  extern char *optarg;
  extern int optind, opterr, optopt;
  
  // parse command line
  int VALIDATE=0, MODE=6, OUTPUT=0, LCP_COMPUTE=0, k=0;
  int Verbose=0, OutputGapLcp=0, OutputBwt=0, Extract=0, len_file=0, c;
  char *c_file=NULL;
  while ((c=getopt(argc, argv, "cm:lvbg:hxXLo")) != -1) {
    switch (c) 
      {
      case 'c':
        VALIDATE=1; break;          // validate output
      case 'm':
        MODE=atoi(optarg); break;   // algorithm to be used
      case 'o':
        OUTPUT=1; break;            // output SA possibly combined with LCP
      case 'l':
        LCP_COMPUTE=1;  break;      // compute LCP 
      case 'v':
        Verbose++; break; 
      case 'x':
        Extract+=1; break;          // extract input sequences and store them in individual file   
      case 'X':
        Extract+=2; break;          // extract input sequences and store them in individual file   
      case 'L':
        len_file=1; break;  // name file containing the length of the concatenated sequences 
      case 'b':
        OutputBwt=1; break;        // output BWT  
      case 'g':
        OutputGapLcp=atoi(optarg); LCP_COMPUTE=1; break;  // compute LCP and output in Gap format 
      case 'h':
        usage(argv[0]); break;       // show usage and stop
      case '?':
        exit(EXIT_FAILURE);
      }
  }
  if(optind+2==argc) {
    c_file=argv[optind++];
    k =atoi(argv[optind++]);
  }
  else  usage(argv[0]);

  //if(MODE>6) LCP_COMPUTE=1;  // not sure about MODE 7: do nothing just allocate? 
  if(MODE<1 || MODE>6) {
    puts("Invalid algorithm!!!\n");
    usage(argv[0]);
  }
  if(OutputGapLcp<0 || OutputGapLcp >= 8) {
    puts("Invalid lcp size!! Must be between 1 and 7\n");
    usage(argv[0]);
  }
  if(OutputGapLcp > sizeof(int_t)) {
    puts("Invalid lcp size!! Set m64 to 1 in lib/utils.h to use 8-byte LCPs\n");
    usage(argv[0]);
  }
    
  // inits 
  time_t t_start=0, t_total=0;
  clock_t c_start=0, c_total=0;
  unsigned char **R;
  int_t i, n=0;
  // print command line 
  printf("Command line:\n ");
  for(i =0; i<argc; i++) printf(" %s",argv[i]);
  puts("");

  // disk access
  if(len_file==0)
    R = (unsigned char**) file_load_multiple(c_file, k, &n);
  else {
    char lenFilename[500]; 
    snprintf(lenFilename,500,"%s.len",c_file);
    R = file_load_concat(c_file,lenFilename,k,&n);
  }
  if(!R){
    fprintf(stderr, "Error: less than %d strings in %s\n", k, c_file);
    return 0;
  }

  if(Extract>0) { // save files and exit
    char s[500]; 
    if(Extract==1) {
      for(i=0;i<k;i++) { // save individual files
        snprintf(s,500,"%s.%"PRIdN,c_file,i);
        FILE *f=fopen(s,"wb");
        size_t r = fwrite((char *)R[i],1,strlen((char *)R[i]),f);
        if(r!=strlen((char *)R[i])) die(__func__);
        fclose(f);
      }  
    }
    else { // save concatenated file 
      snprintf(s,500,"%s.%d.cat",c_file,k);
      FILE *f=fopen(s,"wb");
      snprintf(s,500,"%s.%d.len",c_file,k);
      FILE *len=fopen(s,"wb");
      for(i=0;i<k;i++) {
        uint64_t j = strlen((char *)R[i]);
        #if  __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
          size_t z = fwrite(&j,4,1,len);
          assert(z==1); 
        #else
          fputc(j,len); fputc(j>>8,len); fputc(j>>16,len); fputc(j>>24,len);
        #endif
        size_t r = fwrite((char *)R[i],1,strlen((char *)R[i]),f);
        if(r!=strlen((char *)R[i])) die(__func__);
      }
      fclose(len);
      fclose(f);
    }
    exit(0);
  }
  
  unsigned char *str = NULL;
  int_t  *str_int = NULL;
  //concatenate strings
  if(MODE==1 || MODE==2){//sais or saca-k (int) 
    str_int = cat_int(R, k, &n);
    
    #if DEBUG
    int_t i;
    for(i=0;i<min(10,n); i++)
       printf("%" PRIdN ") %" PRIdN "\n", i, str_int[i]);
    printf("\n");
    #endif
  }
  else{ // sais, saca-k, gsais, gsaca-k (char)
    str = cat_char(R, k, &n);
    #if DEBUG
    int_t i;
    for(i=0;i<min(10,n); i++)
       printf("%" PRIdN ") %d %c\n", i, str[i], str[i]);
    printf("\n");
    #endif
  }

  printf("K = %" PRId32 "\n", k);
  printf("N = %" PRIdN "\n", n+1);
  printf("sizeof(int) = %zu bytes\n", sizeof(int_t));

  #if DEBUG
    printf("R:\n");
    for(i=0; i<min(5,k); i++)
      printf("%" PRIdN ") %s (%zu)\n", i, R[i], strlen((char*)R[i]));
  #endif

  //free memory
  for(i=0; i<k; i++)
    free(R[i]);
  free(R);

  int_t *SA = (int_t*) malloc(n*sizeof(int_t));
  for(i=0; i<n; i++) SA[i]=0;
  int_t depth=0;

  int_t *LCP = NULL;  
  if(LCP_COMPUTE){
    LCP = (int_t*) malloc(n*sizeof(int_t));
    for(i=0; i<n; i++) LCP[i]=0;
  }
  
  time_start(&t_total, &c_total);
  time_start(&t_start, &c_start);

  switch(MODE){
    case 1: printf("## SAIS (int) ##\n");
      depth = SAIS((int_t*)str_int, SA, n, 256+k, sizeof(int_t), 0);
      break;

    case 2: printf("## SACA_K (int) ##\n"); 
      depth = SACA_K((int_t*)str_int, (uint_t*)SA, n, 256+k, n, sizeof(int_t), 0);
      break;
    
    case 3: printf("## SAIS (char) ##\n");
      depth = SAIS((int_t*)str, SA, n, 256, sizeof(char), 0);
      break;

    case 4: printf("## SACA_K (char) ##\n"); 
      depth = SACA_K((int_t*)str, (uint_t*)SA, n, 256, n, sizeof(char), 0);
      break;
  
    case 5: printf("## gSAIS ##\n"); 
      depth = gSAIS((unsigned char*)str, SA, n, 256, sizeof(char), 0, 1);//separator=1
      break;

    case 6: printf("## gSACA_K ##\n"); 
      depth = gSACA_K((unsigned char*)str, (uint_t*)SA, n, 256, sizeof(char), 0, 1);//separator=1
      break;

    default: break;
  }

  if(Verbose)
    fprintf(stderr,"%.6lf\n", time_stop(t_start, c_start));

  // output BWT  
  if(OutputBwt) {
    char s[500]; 
    int c; int_t i;
    snprintf(s,500,"%s.bwt",c_file);
    FILE *f_out = file_open(s, "wb");
    for(i=0; i<n; i++) {
      if(i==0)
        assert(SA[i]==n-1);
      else { 
        if(SA[i]==0 || str[SA[i]-1]==1)
          c = 0;
        else  
          c = str[SA[i]-1]-1;
        int err = fputc(c,f_out);
        if(err==EOF) die(__func__);
      }
    }
    fclose(f_out);
  }


  if(LCP_COMPUTE && MODE<7){
    time_start(&t_start, &c_start);
    if(MODE==1 || MODE==2)
      lcp_PHI_int((int_t*)str_int, SA, LCP, n, sizeof(int_t));
    else
      lcp_PHI((unsigned char*)str, SA, LCP, n, sizeof(char), 1);//separator=1
    printf("PHI-algorithm:\n");
    fprintf(stderr,"%.6lf\n", time_stop(t_start, c_start));
  }
  printf("total:\n");
  fprintf(stderr,"%.6lf\n", time_stop(t_total, c_total));

  if(Verbose>2) {
    if(MODE==1 || MODE==2)//sais or saca-k  
      if(LCP_COMPUTE) lcp_array_print((unsigned char*)str_int, SA, LCP, min(20,n), sizeof(int_t));  
      else suffix_array_print((unsigned char*)str_int, SA, min(10,n), sizeof(int_t)); 
    else
      if(LCP_COMPUTE) lcp_array_print((unsigned char*)str, SA, LCP, min(20,n), sizeof(char)); 
      else suffix_array_print((unsigned char*)str, SA, min(10,n), sizeof(char));
  }


  // validate 
  if(VALIDATE){
    if(MODE==1 || MODE==2){//sais or saca-k 
            if(!suffix_array_check((unsigned char*)str_int, SA, n, sizeof(int_t), 0)) printf("isNotSorted!!\n");
            else printf("isSorted!!\ndepth = %" PRIdN "\n", depth);
    }
    else if(MODE==3 || MODE==4){
            if(!suffix_array_check((unsigned char*)str, SA, n, sizeof(char), 0)) printf("isNotSorted!!\n");//compares until the sentinel=0
            else printf("isSorted!!\ndepth = %" PRIdN "\n", depth);
    }
    else if(MODE==5 || MODE==6){
            if(!suffix_array_check((unsigned char*)str, SA, n, sizeof(char), 1)) printf("isNotSorted!!\n");//compares until the separator=1
            else printf("isSorted!!\ndepth = %" PRIdN "\n", depth);
    }
    if(LCP_COMPUTE){
      if(MODE==1 || MODE==2)//sais or saca-k  
                    if(!lcp_array_check_phi((unsigned char*)str_int, SA, LCP, n, sizeof(int_t), 1)) printf("isNotLCP!!\n");
                    else printf("isLCP!!\n");
      else
                    if(!lcp_array_check_phi((unsigned char*)str, SA, LCP, n, sizeof(char), 1)) printf("isNotLCP!!\n");
                    else printf("isLCP!!\n");
    }
  }
  else if(Verbose) printf("depth = %" PRIdN "\n", depth);

  // output SA alone or SA&LCP together
  if(OUTPUT){
    if(LCP_COMPUTE) lcp_array_write(SA, LCP, n, c_file, "sa_lcp");
    else suffix_array_write(SA, n, c_file, "sa");
  }

  if(OutputGapLcp){
    char s[500]; 
    uint64_t c; int_t i;
    uint64_t lcp_limit = (1LL << (8*OutputGapLcp))-1;
    snprintf(s,500,"%s.%d.lcp",c_file,OutputGapLcp);
    FILE *f_out = file_open(s, "wb");
    for(i=1;i<n;i++) {  
      c = LCP[i];
      if(c>lcp_limit) {
        fprintf(stderr,"   !!! LCP entry larger than %"PRId64"\n", lcp_limit);
        fprintf(stderr,"   !!! Re-run using more bytes per LCP entry. Exiting...\n");
        exit(EXIT_FAILURE);
      }
      fwrite(&c,OutputGapLcp, 1, f_out);
    }
    fclose(f_out);
  }

  free(SA);
  if(LCP_COMPUTE){
    free(LCP);
  }
  if(MODE==1 || MODE==2) free(str_int);
  else free(str);
  return 0;
}
