#include "util.h"
#include "alphabet.h"
#include "mergegap.h"
#include "mergehm.h"
#include "threads.h"

#define min(a,b) ((a)<(b) ? (a) : (b))

void multiround(bool hm, int group_size,char *path, g_data *input, int num_threads)
{
  char filenamein[Filename_size],filenameout[Filename_size];
  FILE *lcpout=NULL;
  customInt offset, tot_symb=0;
  // threads related
  pc_system merge;
  pthread_t t[num_threads];
  g_data buffer[Threads_buf_size];
  int e;
  
  // init threads
  if(num_threads>0) {
    pc_system_init(&merge,Threads_buf_size);
    merge.buffer = (void *) buffer;
    merge.hm = hm;
    for(int i=0;i<num_threads;i++) {
      e = pthread_create(&t[i], NULL, merger, &merge);
      if(e) die("multiround create");
    }
    if(input->verbose>0) printf("%d merger threads created\n",num_threads);
  }
  
  // multiround merging  
  for(int r=0; ;r++) {     // do as many rounds as necessary
    
    if(input->numBwt<= group_size) break;  // we have reached the last round 
  
    // main loop to be executed by parallel threads
    // offset is within the number of bwt
    if(num_threads>0) {
      e = pthread_mutex_lock(&merge.remutex);
      if(e) die("remaining lock 1");
      merge.remaining = (input->numBwt+group_size-1)/group_size; // parallel merges in this round
      e= pthread_mutex_unlock(&merge.remutex);
      if(e) die("remaining unlock 1");
    }
    tot_symb = 0; 
    for(offset=0; offset < input->numBwt; offset += group_size) {
      g_data g=*input; // copy current status
      if(g.verbose>2) printf("Round %d, NumBwt offset "CUSTOM_FORMAT"\n",r,offset);
      g.bws = input->bws+offset;
      g.bwtLen = input->bwtLen+offset;
      if(input->smallAlpha) g.bwtOcc = input->bwtOcc+offset;
      g.numBwt = min(group_size, input->numBwt - offset);
      g.symb_offset = tot_symb;
      g.mergeLen=0;
      for(int i=0;i<g.numBwt;i++) g.mergeLen += g.bwtLen[i];
      // prepare access to lcp values
      if(!g.bwtOnly) {
        g.blockBeginsAt = input->blockBeginsAt + g.symb_offset;
        if(g.lcpMem) g.lcps = input->lcps+offset;
        else assert(g.singleFile);
      }
      // execute merge, possibly using threads 
      if(num_threads>0) {
        e = sem_wait(&merge.free_slots); if(e) die("multiround producer wait");
        check_g_data(&g);
        buffer[merge.pindex++ % merge.buf_size]=g; // copy g_data to buffer
        e = sem_post(&merge.ready);  if(e) die("multiround producer post");
      }
      else { // no helper threads
        if (hm) holtMcMillan(&g, false);
        else gap(&g, false);
        // update input via g
        for(int i=1;i<g.numBwt;i++) {
          g.bwtLen[0] += g.bwtLen[i];
          if(g.smallAlpha)
            for(int j=0;j<g.sizeOfAlpha;j++)
              g.bwtOcc[0][j] += g.bwtOcc[i][j];
        }
        assert(g.bwtLen[0]==g.mergeLen);
      }
      // update tot_symb seen so far
      tot_symb += g.mergeLen;
    } // end of inner parallelizable loop
    assert(input->mergeLen==tot_symb);
    // --- wait all merges for this round have been completed 
    if(num_threads>0) {
      e = pthread_mutex_lock(&merge.remutex); if(e) die("remaining lock 2");
      while(merge.remaining>0) {
        e = pthread_cond_wait(&merge.recond,&merge.remutex);
        if(e) die("remaining wait");
      }
      e = pthread_mutex_unlock(&merge.remutex); if(e) die("remaining unlock 2");
      assert(merge.pindex==merge.cindex);
      if(input->verbose>1) puts("Threads syncronized, round completed");
    }
    
    // merged done for all segments. merged bwt/lcp are in g.bws[0][] g.blockBeginsAt[]
    if(!input->bwtOnly) {
      // save new lcp values
      if(!input->lcpMem) { // save to temporary file
        snprintf(filenameout,Filename_size,"%s.%d.tmp.%s",path,r,LCP_EXT);
        lcpout = fopen(filenameout,"wb");
        if(lcpout==NULL) die("Cannot open temporary lcp file for writing");
        int w = fwrite(input->blockBeginsAt,sizeof(lcpInt),tot_symb,lcpout);
        if(w!=tot_symb) die(__func__);
        fclose(lcpout);
        // save base name for reading lcp at next round
        snprintf(filenamein,Filename_size,"%s.%d.tmp",path,r);
        input->lcpinPath = filenamein;
      } 
      else  // save to lcps[0]
        memcpy(input->lcps[0],input->blockBeginsAt,tot_symb*sizeof(lcpInt));
    }
    // update input so that it consists of fewer, larger segments. move Len and Occ real entries to the left
    for(offset=0; offset < input->numBwt; offset += group_size) {
      input->bwtLen[offset/group_size] = input->bwtLen[offset];
      input->bws[offset/group_size] = input->bws[offset];
      if(input->lcpMem) input->lcps[offset/group_size] = input->lcps[offset];;
      // replace with a swap ??? 
      if(input->smallAlpha)
        for(int j=0;j<input->sizeOfAlpha;j++)
          input->bwtOcc[offset/group_size][j] = input->bwtOcc[offset][j];
    }
    input->numBwt = (input->numBwt+group_size-1)/group_size;
    check_g_data(input);
    if(input->verbose>0)  printf("Round %d complete: %d merged bwt/lcp saved\n",r,input->numBwt);
  } // end of all rounds except last
  if(num_threads>0) {  // last round does not use threads
    g_data g; // dummy task
    g.numBwt = 0;
    g.verbose = input->verbose;
    for(int i=0;i<num_threads;i++) {
        e = sem_wait(&merge.free_slots); if(e) die("cleaning producer wait");
        buffer[merge.pindex++ % merge.buf_size]=g; // copy g_data to buffer
        e = sem_post(&merge.ready);  if(e) die("cleaning producer post");
    }
    for(int i=0;i<num_threads;i++) {
      e = pthread_join(t[i], NULL);
      if(e) die("multiround join");
    }
    pc_system_destroy(&merge);
    if(input->verbose>1) printf("%d merge threads destroyed",num_threads);
  }
  // execute last round
  
  if (hm) holtMcMillan(input, true);
  else gap(input, true);
  
}

// remap BWTs optionally squeeze them, and init g>bwtOcc[i] and g->sizeOfAlpha
void remap_bwts(g_data *g, bool squeeze, int num_threads)
{
  int sizePreAlphabet,sizeOfAlphabet; // size of alphabet before and after squeezing
  // threads related
  pc_system sqz;
  pthread_t t[num_threads];
  int buffer[Threads_buf_size];
  int e;


  // remap alphabet, remap bwts and compute occ 
  sizePreAlphabet = init_alpha_maps(g->bws[0], g->mergeLen);
  assert(sizePreAlphabet>1 && sizePreAlphabet<= SIZE_OF_ALPHABET);
  // shall we squeeze? 
  g->charXsymb = squeeze ? squeezable(sizePreAlphabet) : 1;
  if(g->charXsymb==1)
    sizeOfAlphabet=sizePreAlphabet;
  else 
    sizeOfAlphabet = intpow(sizePreAlphabet,g->charXsymb); // using ktuples: set new alphabet size
  // init some globals related to squeezing, works fine even if no squeezing   
  ktuple_init(sizePreAlphabet,g->charXsymb);
  if(g->verbose>0) ktuple_info(stdout);    
  
  // if requested create threads
  if(g->charXsymb>1 && num_threads>0) {
    g->sizeOfAlpha = sizePreAlphabet;  // need to be passed to threads
    pc_system_init(&sqz,Threads_buf_size);
    sqz.buffer = (void *) buffer;
    sqz.data = g;
    for(int i=0;i<num_threads;i++) {
      e = pthread_create(&t[i], NULL, squeezer, &sqz); 
      if(e) die("remap_bwts create");
    }
    if(g->verbose>0) printf("%d squeezer threads created\n",num_threads);
  }          
  
  // remap and, if requested, squeeze. also compute Occs for each BWT
  if(g->smallAlpha) {
    g->bwtOcc = malloc(g->numBwt*sizeof(customInt *));
    if(!g->bwtOcc) die(__func__);
    g->bwtOcc[0] = calloc(g->numBwt*sizeOfAlphabet,sizeof(customInt));
   if(!g->bwtOcc[0]) die(__func__);
  }  
  for(int i=0;i<g->numBwt;i++) {
    customInt *bwtOcc = NULL;
    if(g->smallAlpha) { // if the alphabet is small keep bwtOcc[i]
      if(i>0) g->bwtOcc[i] = g->bwtOcc[i-1] + sizeOfAlphabet;
      bwtOcc = g->bwtOcc[i];
    }
    if(g->charXsymb==1) // no squeezing just remap and compute g->bwtOcc[i] 
      remap_string(g->bws[i],g->bwtLen[i],bwtOcc);
    else {
      if(num_threads>0) {
        e = sem_wait(&sqz.free_slots); if(e) die("remap_bwts producer wait");
        buffer[sqz.pindex++ % sqz.buf_size]=i;  // copy g_data to buffer
        e = sem_post(&sqz.ready);  if(e) die("remap_bwts producer post");        
      }
      else
        ktuple_remap_string(g->bws[i],g->bwtLen[i],bwtOcc,sizePreAlphabet,g->charXsymb);
    }
  }

  g->sizeOfAlpha = sizeOfAlphabet;

  // destroy threads and related structures 
  if(g->charXsymb>1 && num_threads>0) {
    for(int i=0;i<num_threads;i++) {
        e = sem_wait(&sqz.free_slots); if(e) die("cleaning producer wait");
        buffer[sqz.pindex++ % sqz.buf_size]= -1;   // invalid index 
        e = sem_post(&sqz.ready);  if(e) die("cleaning producer post");
    }
    for(int i=0;i<num_threads;i++) {
      e = pthread_join(t[i], NULL);
      if(e) die("remap_bwts join");
    }
    pc_system_destroy(&sqz);
    if(g->verbose>1) printf("%d squeezer threads destroyed",num_threads);
  }          

}

