#include "util.h"
#include "alphabet.h"

#define min(a,b) ((a)<(b) ? (a) : (b))
#define minLen 40


void array_clear(customInt* array, customInt size, customInt v) {
  for (customInt i = 0; i < size; ++i ) array[i] = v;
}


/**
 * read BWTs from a single file
 * */
customInt readBWTsingle(char *path, g_data *g) // ; int num, symbol **bwt, customInt *bwtsize)
{
  char filename[Filename_size];
  customInt n=0;

  // open and read .len file
  snprintf(filename,Filename_size,"%s.%s",path,LEN_EXT);
  FILE *f = fopen(filename,"r");
  if(f==NULL) die(__func__);
  for (int i = 0; i < g->numBwt; ++i) {
    uint32_t size;
    size_t r = fread(&size, 4, 1, f);
    if(r!=1) {
      fprintf(stderr,"Found only %d bwt-lengths in file %s. Expected %d\n",
      i,filename,g->numBwt);
      die(__func__);
    }
    if(size>MAX_OUTPUT_SIZE) {
      fprintf(stderr,"Bwt len %"PRIu32" larger than maximum sequence size: %llu\n",
      size,MAX_OUTPUT_SIZE);
      die(__func__);
    }
    g->bwtLen[i] = size;
    n += size;
  }
  fclose(f);

  // open concatenated bwt file 
  snprintf(filename,Filename_size,"%s.%s",path,BWT_EXT);
  int fd = open(filename,O_RDONLY);
  if(fd == -1) die(__func__);    
  // allocate memory for BWT concatenation and read whole file
  if(g->mmapBWT) {
    g->bws[0] = mmap(NULL,n*sizeof(symbol),PROT_READ|PROT_WRITE,MAP_PRIVATE,fd,0);
    if(g->bws[0] == MAP_FAILED) die(__func__);
  }
  else { // allocate and read from file descriptor
    uint8_t *b = malloc(n*sizeof(symbol));
    if(!b) die(__func__);  
    size_t tot=0;
    do {
      ssize_t z = read(fd, b + tot,(sizeof(symbol)*n)-tot);
      if(z<0) {perror("BWT read in readBWTsingle"); die(__func__);}
      tot += z;
    } while(tot < sizeof(symbol)*n);
    g->bws[0] = (symbol *) b;
  } 
  if(close(fd)!=0) die(__func__);
  
  // init starting positions 
  for (int i = 0; i < g->numBwt -1; ++i)  
    g->bws[i+1] = g->bws[i] + g->bwtLen[i];
  return n; 
}

/**
 * read num bwt's and store them in memory
 * 
 * /param path  base file name
 * /param num   number of bwt's to be read
 * /return      sum of size of bwt's
 * 
 * Using realloc the bwt's are stored in a single array of symbols
 * whose size is returned (and becomes sizeOfMerge)
 * Init the, already allocated, arrays 
 *   bwt[num]     starting point of each bwt
 *   bwtsize[num] size of each bwt
 * 
 * Note: the fact that bwt's are stored in the same array is fundamental 
 *       for multiround merge
 * */ 
customInt readBWTfiles(char *path, int num, symbol **bwt, customInt *bwtsize)
{
  char filename[Filename_size];
  symbol *base = NULL;
  customInt n=0;
  
  for (int i = 0; i < num; ++i) {
    // create filename and open
    snprintf(filename,Filename_size,"%s.%d.%s",path,i,BWT_EXT);
    FILE *file = fopen(filename,"r");
    if(file==NULL) die(__func__);  
    // get file size and allocate 
    if(fseek(file, 0, SEEK_END) != 0) die(__func__);
    off_t size = ftello(file);
    if(size==0) {fprintf(stderr,"Zero length file: %s\n", filename);die(__func__);}
    if(size>MAX_OUTPUT_SIZE) {fprintf(stderr,"File too large: %s\n", filename);die(__func__);}
    base = realloc(base,(n+size)*sizeof(symbol));
    if(base==NULL) die(__func__);
    // rewind and read
    rewind(file);
    size_t r =  fread(base+n,sizeof(symbol),size,file);
    if(r!=size) die(__func__);
    fclose(file); 
    // update sizes
    bwtsize[i] = size;
    n += size;
  }
  // init bwt[]
  bwt[0] = base;
  for (int i = 0; i < num -1; ++i)  
    bwt[i+1] = bwt[i] + bwtsize[i];
  return n;  
}

void array_copy(customInt *dest, customInt *src, customInt size){
  for (customInt i = 0; i < size; ++i) {
    dest[i] = src[i];
  }
}

void testPrintMergeColor(customInt prefixLength, palette *mergeCol, customInt size) {    
        printf("Z("CUSTOM_FORMAT"):\n",prefixLength);
        for (customInt i = 0; i < min(minLen,size); i++)
            printf(PALETTE_FORMAT" ", mergeCol[i]);
  printf("\n");
}

void testPrintI(char *name, customInt param, customInt *array, customInt size) {
  printf("%s("CUSTOM_FORMAT"):\n",name,param);
  for (customInt i = 0; i < min(minLen,size); i++)
    printf(CUSTOM_FORMAT" ",array[i]);        
  printf("\n");
}

void testPrintLcp(char *name, customInt param, lcpInt *array, customInt size) {
  printf("%s("CUSTOM_FORMAT"):\n",name,param);
  for (customInt i = 0; i < min(minLen,size); i++)
    printf(LCP_FORMAT" ",array[i]);        
  printf("\n");
}


// debug only funcion 
void check_g_data(g_data *g)
{
  for(int i=0;i<g->numBwt-1;i++) {
    assert(g->bws[i]+g->bwtLen[i]==g->bws[i+1]);
  }
  if(g->smallAlpha) 
    for(int i=0;i<g->numBwt;i++) {
      int tot=0;
      for(int j=0;j<g->sizeOfAlpha;j++) {
        tot += g->bwtOcc[i][j];
      }
      assert(tot==g->bwtLen[i]);
    }
  customInt mergeLen=0;
  for(int i=0;i<g->numBwt;i++) mergeLen += g->bwtLen[i];
  assert(mergeLen==g->mergeLen);
}

  
// create g->numBwt file pointers and store them in files[]
// each one pointing to the beginning of a lcp file
// this function do not support MultiRound since file index starts with  0  
void lcp_files_open_multi(g_data *g, char *path, FILE *files[])
{
  char filename[Filename_size];
  assert(g->symb_offset==0); // function not to be used in multiround, offset not suppoorted
  for(int i=0;i<g->numBwt;i++) {
    snprintf(filename,Filename_size,"%s.%d.%s",path,i,LCP_EXT);
    if(g->verbose>2) printf("Using LCP file %s\n",filename);
    files[i] = fopen(filename,"rb");
    if(files[i]==NULL) die(__func__);
  }
}



// if files!=NULL create g->numBwt file pointers and store them in files[]
// each one pointing to the beginning of a different lcp portion
void lcp_files_open_single(g_data *g, char *path, FILE *files[])
{
  char filename[Filename_size];
  snprintf(filename,Filename_size,"%s.%s",path,LCP_EXT);
  if(g->verbose>2) printf("Using LCP file %s\n",filename);
  long offset = g->symb_offset;
  for(int i=0;i<g->numBwt;i++) {
    files[i] = fopen(filename,"rb");
    if(files[i]==NULL) die(__func__);
    int h = fseek(files[i],offset*sizeof(lcpInt),SEEK_SET);
    if(h == -1) die(__func__);
    offset += g->bwtLen[i]; 
  }
}


#if 0
// === read lcp values from file 

// read lcp values into memory g->lcps must be non-null and the lcp values
// from path are stored in the correct portion of g->lcps
void lcp_files_read_single(g_data *g)
{
  char filename[Filename_size];

  // open file and go to the correct position 
  snprintf(filename,Filename_size,"%s.%s",g->lcpinPath,LCP_EXT);
  if(g->verbose>2) printf("Using LCP file %s\n",filename);
  FILE *f = fopen(filename,"rb");
  if(f==NULL) die(__func__);
  // read lcp values 
  size_t z = fread(g->lcps[0],sizeof(lcpInt),g->mergeLen,f);
  if(z!=g->mergeLen) {
    fprintf(stderr,"Invalid LCP file %s (%zu vs "CUSTOM_FORMAT")\n",filename,z,g->mergeLen);
    die(__func__);
  }
  fclose(f); // close file
}

void lcp_files_read_multi(g_data *g)
{
  char filename[Filename_size];
  assert(g->lcps!=NULL);
  
  for(int i=0;i<g->numBwt;i++) {
    snprintf(filename,Filename_size,"%s.%d.%s",g->lcpinPath,i,LCP_EXT);
    if(g->verbose>2) printf("Using LCP file %s\n",filename);
    FILE *f = fopen(filename,"rb");
    if(f==NULL) die(__func__);
    size_t z = fread(&(g->lcps[i]),sizeof(lcpInt),g->bwtLen[i],f);
    if(z!=g->bwtLen[i]) {
       fprintf(stderr,"Invalid LCP file %s (%zu vs "CUSTOM_FORMAT")\n",filename,z,g->bwtLen[i]);
      die(__func__);
    }
    fclose(f); // close file
  }
}
#endif


// allocate and fill the LCP values in memory (mmapped)
void initLCPmem(g_data *g)
{
  char filename[Filename_size];

  assert(g->singleFile);
  // allocate g->lcps and g->lcp[0..]
  g->lcps = (lcpInt **) malloc(g->numBwt*sizeof(lcpInt *));
  if(!g->lcps) die(__func__);
  
  snprintf(filename,Filename_size,"%s.%s",g->lcpinPath,LCP_EXT);
  int fd = open(filename,O_RDONLY);
  if(fd == -1) die(__func__);
  // mmap lcp values
  g->lcps[0] = mmap(NULL,g->mergeLen*sizeof(lcpInt),PROT_READ|PROT_WRITE,MAP_PRIVATE,fd,0);
  if(g->lcps[0] == MAP_FAILED) die(__func__);
  if(close(fd)!=0) die(__func__);
  // init lcps[i..]
  for(int i=0;i<g->numBwt-1;i++)
    g->lcps[i+1] = g->lcps[i] + g->bwtLen[i];
}
    


/**
 * function to merge the BWTs, and optionally LCPs, according to 
 * g->mergeColor[]. LCP values g->blockBeginsAt[]
 * \param bwtout[] where the merged BWT is written, it is casted mergeColor
 * \param lcpin[]  file pointers to the input lcp values 
 * \param g        contains: blockBeginsAt[] mergeColor[] bws[][]
 *                 inCnt[] mergeLen numBwt
 * */
void mergeBWTandLCPaux(symbol *bwtout, FILE *lcpin[], g_data *g)
{
  // bwtout is mergeColor
  check_g_data(g);
  array_clear(g->inCnt,g->numBwt,0);
  for (customInt i = 0; i < g->mergeLen; ++i) {
    int currentColor = g->mergeColor[i];
    assert(currentColor < g->numBwt);
    // write new BWT char
    symbol c = g->bws[currentColor][g->inCnt[currentColor]];
    bwtout[i] = c;
    if(!g->bwtOnly) { // compute lcp value
      lcpInt tmp=255;
      if(g->lcpMem)  // read from g->lcps
        tmp = g->lcps[currentColor][g->inCnt[currentColor]]; // lcp retrieved from file (suffixes from the same string)
      else {            // read from file
        int r = fread(&tmp, sizeof(tmp), 1, lcpin[currentColor]);
        if(r!=1) {
          fprintf(stderr,"Error reading at position "CUSTOM_FORMAT" (items read %d)\n",i,r);
          die(__func__);
        }
      }
      if(g->blockBeginsAt[i]>0)
        g->blockBeginsAt[i] -= 1;  // lcp computed by the merging algorithm 
      else 
        g->blockBeginsAt[i] = tmp; // lcp retrieved from file (suffixes from the same string) 
    }
    g->inCnt[currentColor]++; // one more char read from currentColor BWT/LCP
  }
  for(int i=0;i<g->numBwt;i++) assert(g->inCnt[i]==g->bwtLen[i]);
}

// uses g->mergeColor to merge bwt values from g->bws[i][j] to g->newMergeColor
// and possibly transforms g->blockBeginsAt in a true LCP value  
// Note that we are opening g->numBwt files this will cause an immediate heap usage
// of 576 bytes per file (and maybe more for file buffers) this can be a significant 
// overhead for small inputs. Cosider use mmap instead.     
void mergeBWTandLCP(g_data *g)
{
  if(sizeof(palette) < sizeof(symbol)) die("Sorry, alphabet is too large wrt Max # BWTs.");
  FILE **lcpin = NULL;

  if(!g->bwtOnly && !g->lcpMem) {
    lcpin = malloc(g->numBwt*sizeof(FILE *));
    if(!lcpin) die(__func__);
    // open lcpfiles  
    if(g->singleFile)
      lcp_files_open_single(g,g->lcpinPath,lcpin);
    else 
      lcp_files_open_multi(g,g->lcpinPath,lcpin);
  }
  // do the actual merge 
  mergeBWTandLCPaux((symbol *) g->mergeColor, lcpin, g);
  // close lcpin[]
  if(!g->bwtOnly && !g->lcpMem) {
    for(int i=0;i<g->numBwt;i++)
      fclose(lcpin[i]);
    free(lcpin);
  }
}

// allocate or mmap arrays Z and newZ
void alloc_merge_arrays(g_data *g) {
  if(g->mmapZ) {
    g->mergeColor = mmap(NULL,g->mergeLen*sizeof(palette),PROT_READ|PROT_WRITE,MAP_PRIVATE|MAP_ANONYMOUS,-1,0);
    if(g->mergeColor== MAP_FAILED) die(__func__);
    g->newMergeColor =  mmap(NULL,g->mergeLen*sizeof(palette),PROT_READ|PROT_WRITE,MAP_PRIVATE|MAP_ANONYMOUS,-1,0);
    if(g->newMergeColor== MAP_FAILED) die(__func__);
  }
  else {
    g->mergeColor =     malloc(g->mergeLen*sizeof(palette)); 
    g->newMergeColor =  malloc(g->mergeLen*sizeof(palette));
     if(!g->mergeColor || !g->newMergeColor) die(__func__);
  }
}

void free_merge_arrays(g_data *g) {
  if(g->mmapZ) {
    int e = munmap(g->mergeColor,g->mergeLen*sizeof(palette));
    if(e!=0) die("gap (unmap mergeColor)");
    e = munmap(g->newMergeColor,g->mergeLen*sizeof(palette));
    if(e!=0) die("gap (unmap newMergeColor)");  }
  else {
    free(g->newMergeColor);
    free(g->mergeColor);
  }
}


// allocate or mmap array Z and not newZ
void alloc_merge_array(g_data *g) {
  if(g->mmapZ) {
    g->mergeColor = mmap(NULL,g->mergeLen*sizeof(palette),PROT_READ|PROT_WRITE,MAP_PRIVATE|MAP_ANONYMOUS,-1,0);
    if(g->mergeColor== MAP_FAILED) die(__func__);
  }
  else {
    g->mergeColor =     malloc(g->mergeLen*sizeof(palette)); 
     if(!g->mergeColor) die(__func__);
  }
  g->newMergeColor = NULL;
}

void free_merge_array(g_data *g) {
  assert(g->newMergeColor==NULL);
  if(g->mmapZ) {
    int e = munmap(g->mergeColor,g->mergeLen*sizeof(palette));
    if(e!=0) die("gap (unmap mergeColor)");
  }
  else {
    free(g->mergeColor);
  }
}

 
// error message and exit
void die(const char* where)  {
  printf("Error at %s: %s.\n",where,errno ? strerror(errno) : "errno not set");
  exit(errno?errno:1);
}
